import os

class UploadDiskDriver(object):

    def __init__(self, StorageConfig, Application):
        self.config = StorageConfig
        self.appconfig = Application

    def store(self, fileitem, location=None):
        filename = os.path.basename(fileitem.filename)

        if not location:
            location = self.config.DRIVERS['disk']['location']
        
        open(location + filename, 'wb').write(fileitem.file.read())

        return location + filename

    def storePrepend(self, fileitem, prepend, location=None):
        filename = os.path.basename(fileitem.filename)

        if not location:
            location = self.config.DRIVERS['disk']['location']

        open(location + prepend + filename, 'wb').write(fileitem.file.read())

        return location + prepend + filename



''' Request Module handles many different aspects of a single request
    Methods which require the request and help ease development should
    be put here.

    Methods may return another object if necessary to expand capabilities
    of this class.
'''
from urllib.parse import parse_qs
from http import cookies
from masonite.auth.Sign import Sign

class Request(object):
    ''' Handles many different aspects of a single request.
        This is the object passed through to the controllers
        as a request paramter
    '''
    def __init__(self, environ):
        self.cookies = []
        self.url_params = None
        self.redirect_url = False
        self.redirect_route = False
        self.user_model = None
        self.environ = environ
        self.params = environ['QUERY_STRING']
        self.method = environ['REQUEST_METHOD']
        self.path = environ['PATH_INFO']
        self.encryption_key = False
        self.container = None

    def input(self, param):
        ''' Returns either the FORM_PARAMS during a POST request
            or QUERY_STRING during a GET request
        '''

        # Post Request Input
        if self.is_post():
            if isinstance(self.params, str):
                return parse_qs(self.params)[param][0]

            if not self.params[param].filename:
                return self.params[param].value

            if self.params[param].filename:
                return self.params[param]

        # Get Request Input
        if self.has(param):
            return parse_qs(self.params)[param][0]

        return False

    def file(self, param):
        pass

    def is_post(self):
        if self.environ['REQUEST_METHOD'] == 'POST':
            return True

        return False

    def key(self, key):
        ''' Sets encryption key '''
        self.encryption_key = key
        return self

    def all(self):
        ''' Returns all the params '''
        if isinstance(self.params, str):
            return parse_qs(self.params)
        
        return self.params

    def load_app(self, app):
        self.container = app
        return self

    def app(self):
        return self.container
    
    def has(self, param):
        ''' Check if a param exists '''
        if param in self.params:
            return True

        return False

    def set_params(self, params):
        ''' Loads the params into the class.
            These parameters are where the developer can retrieve the
            /url/@variable:string/ from the url.
        '''
        self.url_params = params
        return self

    def param(self, parameter):
        ''' Retrieves the param from the URL.
            The "parameter" parameter in this method should be the name of the @variable
            passed into the url in web.py '''
        if parameter in self.url_params:
            return self.url_params[parameter]
        return False

    def cookie(self, key, value, encrypt = True):
        ''' Sets a cookie in the browser '''
        if encrypt:
            value = Sign(self.encryption_key).sign(value)
        else:
            value = value
            
        self.cookies.append(('Set-Cookie', '{0}={1}; HttpOnly'.format(key, value)))
        self.append_cookie(key, value)
        return self

    def get_cookies(self):
        ''' Retrieve all cookies from the browser '''
        return self.cookies

    def get_cookie(self, provided_cookie, decrypt=True):
        ''' Retrieves a specific cookie from the browser '''
        if 'HTTP_COOKIE' in self.environ:
            grab_cookie = cookies.SimpleCookie(self.environ['HTTP_COOKIE'])
            if provided_cookie in grab_cookie:
                if decrypt:
                    return Sign(self.encryption_key).unsign(
                        grab_cookie[provided_cookie].value)
                return grab_cookie[provided_cookie].value

        return None

    def append_cookie(self, key, value):
        if 'HTTP_COOKIE' in self.environ and self.environ['HTTP_COOKIE']:
            self.environ['HTTP_COOKIE'] += ';{0}={1}'.format(
                key, value)
        else:
            self.environ['HTTP_COOKIE'] = '{0}={1}'.format(
                key, value)

    def set_user(self, user_model):
        ''' Loads the user into the class '''
        self.user_model = user_model
        return self

    def user(self):
        ''' Retreives the user model '''
        return self.user_model

    def redirect(self, route):
        ''' Redirect the user based on the route specified '''
        self.redirect_url = route
        return self

    def redirectTo(self, route):
        ''' Redirect to a named route '''
        self.redirect_route = route
        return self

    def back(self, input_parameter='back'):
        ''' Go to a named route with the back parameter '''
        self.redirectTo(self.input(input_parameter))
        return self

    def compile_route_to_url(self):
        ''' Compile the route url into a usable url
            Converts /url/@id into /url/1. Used for redirection
        '''

        if 'http' in self.redirect_url:
            return self.redirect_url

        # Split the url into a list
        split_url = self.redirect_url.split('/')

        # Start beginning of the new compiled url
        compiled_url = '/'

        # Iterate over the list
        for url in split_url:

            # if the url contains a parameter variable like @id:int
            if '@' in url:
                url = url.replace('@', '').replace(':int', '').replace(':string', '')
                compiled_url += '/' + str(self.param(url))
            else:
                compiled_url += url

        return compiled_url

    def send(self, params):
        ''' With '''
        self.set_params(params)
        return self



''' A RouteProvider Service Provider '''
from masonite.provider import ServiceProvider
import re
from pydoc import locate

class RouteProvider(ServiceProvider):

    def register(self):
        pass

    def boot(self, WebRoutes, Route, Request, Environ):
        for route in WebRoutes:
            router = Route
            request = Request
            # Compiles the given route to regex
            regex = router.compile_route_to_regex(route)

            '''
            |--------------------------------------------------------------------------
            | Make a better match for trailing slashes
            |--------------------------------------------------------------------------
            |
            | Sometimes a user will end with a trailing slash. Because the user might
            | create routes like `/url/route` and `/url/route/` and how the regex 
            | is compiled down, we may need to adjust for urls that end or dont 
            | end with a trailing slash.
            |
            '''

            if route.route_url.endswith('/'):
                matchurl = re.compile(regex.replace(r'\/\/$', r'\/$'))
            else:
                matchurl = re.compile(regex.replace(r'\/$', r'$'))

            # This will create a dictionary of parameters given. This is sort of a short
            #     but complex way to retrieve the url parameters. This is the code used to
            #     convert /url/@firstname/@lastname to {'firstmane': 'joseph', 'lastname': 'mancuso'}
            try:
                parameter_dict = {}
                for index, value in enumerate(matchurl.match(router.url).groups()):
                    parameter_dict[router.generated_url_list()[index]] = value
                request.set_params(parameter_dict)
            except AttributeError:
                pass

            '''
            |--------------------------------------------------------------------------
            | Houston, we've got a match
            |--------------------------------------------------------------------------
            |
            | Check to see if a route matches the corresponding router url. If a match
            | is found, execute that route and break out of the loop. We only need
            | one match. Routes are executed on a first come, first serve basis
            |
            '''

            if matchurl.match(router.url) and route.method_type == Environ['REQUEST_METHOD']:

                '''
                |--------------------------------------------------------------------------
                | Execute Before Middleware
                |--------------------------------------------------------------------------
                |
                | This is middleware that contains a before method.
                |
                '''

                for http_middleware in self.app.make('HttpMiddleware'):
                    located_middleware = locate(http_middleware)
                    if hasattr(located_middleware, 'before'):
                        self.app.resolve(located_middleware().before)

                # Show a helper in the terminal of which route has been visited
                print(route.method_type + ' Route: ' + router.url)

                # Loads the request in so the middleware specified is able to use the
                #     request object. This is before middleware and is ran before the request
                route.load_request(request).run_middleware('before')

                # Get the data from the route. This data is typically the output
                #     of the controller method
                if not request.redirect_url:
                    self.app.bind('Response', router.get(route.route, self.app.resolve(route.output)))

                # Loads the request in so the middleware specified is able to use the
                #     request object. This is after middleware and is ran after the request
                route.load_request(request).run_middleware('after')

                '''
                |--------------------------------------------------------------------------
                | Execute After Middleware
                |--------------------------------------------------------------------------
                |
                | This is middleware with an after method.
                |
                '''

                for http_middleware in self.app.make('HttpMiddleware'):
                    located_middleware = locate(http_middleware)
                    if hasattr(located_middleware, 'after'):
                        self.app.resolve(located_middleware().after)
                break
            else:
                data = 'Route not found. Error 404'



from masonite.exceptions import DriverNotFound
from masonite.managers.Manager import Manager

class MailManager(Manager):

    def create_driver(self, driver=None):
        if not driver:
            driver = self.container.make('MailConfig').DRIVER.capitalize()
        else:
            driver = driver.capitalize()

        try:
            self.manage_driver = self.container.make('Mail{0}Driver'.format(driver))
        except KeyError:
            raise DriverNotFound('Could not find the Mail{0}Driver from the service container. Are you missing a service provider?'.format(driver))



''' A Mail Service Provider '''
from masonite.provider import ServiceProvider
from masonite.managers.UploadManager import UploadManager
from masonite.drivers.UploadDiskDriver import UploadDiskDriver
from masonite.drivers.UploadS3Driver import UploadS3Driver
from config import storage

class UploadProvider(ServiceProvider):

    wsgi = False

    def register(self):
        self.app.bind('StorageConfig', storage)
        self.app.bind('UploadDiskDriver', UploadDiskDriver)
        self.app.bind('UploadS3Driver', UploadS3Driver)

    def boot(self):
        self.app.bind('Upload', UploadManager(self.app))



from cryptography.fernet import Fernet

class Sign(object):

    def __init__(self, key=None):
        if key:
            self.key = key
        else:
            from config import application
            self.key = application.KEY

        self.encryption = None

    def sign(self, value):
        f = Fernet(self.key)
        self.encryption = f.encrypt(bytes(value, 'utf-8'))
        return self.encryption.decode('utf-8')
    
    def unsign(self, value=None):
        f = Fernet(self.key)

        if not value:
            return f.decrypt(self.encryption).decode('utf-8')
        return f.decrypt(bytes(value, 'utf-8')).decode('utf-8')



''' A RedirectionProvider Service Provider '''
from masonite.provider import ServiceProvider

class RedirectionProvider(ServiceProvider):

    def register(self):
        pass

    def boot(self, Request, WebRoutes):
        if Request.redirect_route:
            for route in WebRoutes:
                if route.named_route == Request.redirect_route:
                    print(route.method_type + ' Named Route: ' + router.url)
                    data = router.get(route.route, route.output(Request))
                    Request.redirect_url = route.route_url



from masonite.managers.Manager import Manager
from masonite.exceptions import DriverNotFound

class UploadManager(Manager):

    def create_driver(self, driver=None):
        if not driver:
            driver = self.container.make('StorageConfig').DRIVER.capitalize()
        else:
            driver = driver.capitalize()

        try:
            self.manage_driver = self.container.make(
                'Upload{0}Driver'.format(driver))
        except KeyError:
            raise DriverNotFound(
                'Could not find the Upload{0}Driver from the service container. Are you missing a service provider?'.format(driver))



''' Masonite Validator Module '''

from validator import *

class Validator(object):
    ''' Validator Class '''
    def __init__(self, request = None):
        self.request = request
        self.validation_dictionary = {}
        self.check_manual_dictionary = False
        self.error_messages = {}

    def validate(self, dictionary):
        ''' Sets the validation dictionary '''
        self.validation_dictionary = dictionary
        return self

    def check(self, check_manual_dictionary = False):
        ''' Validated the dictionary '''
        if check_manual_dictionary:
            self.check_manual_dictionary = check_manual_dictionary
        return self.run_validation()[0]

    def errors(self):
        ''' Returns Errors '''
        validation = self.run_validation()
        for message in self.error_messages:
            if message in validation[1]:
                validation[1][message] = self.error_messages[message]

        if not validation[1]:
            return None
        
        return validation[1]

    def run_validation(self):
        ''' Loads the dictionary and runs the validations '''
        self.load_request_input()
        return validate(self.validation_dictionary, self.load_request_input())

    def load_request_input(self):
        ''' Need to load request input into a different value
            Since the Request class stores input values as string, we
                need to create a new dictonary value from the request
        '''
        if self.request:
            dictionary = {}
            for value in self.request.all():
                dictionary[value] = self.request.input(value)
            return dictionary

        return self.check_manual_dictionary

    def messages(self, messages):
        ''' You may change the default messages here '''
        self.error_messages = messages



''' A StartResponseProvider Service Provider '''
from masonite.provider import ServiceProvider

class StartResponseProvider(ServiceProvider):

    def register(self):
        pass

    def boot(self, Request, Response):
        if not Request.redirect_url:
            # Convert the data that is retrieved above to bytes so the wsgi server can handle it.

            data = bytes(Response, 'utf-8')

            self.app.bind('StatusCode', "200 OK")
            self.app.bind('Headers', [
                ("Content-Type", "text/html; charset=utf-8"),
                ("Content-Length", str(len(data)))
            ] + Request.get_cookies())
        else:
            self.app.bind('StatusCode', "302 OK")
            self.app.bind('Headers', [
                ('Location', Request.compile_route_to_url())
            ] + Request.get_cookies())
            self.app.bind('Response', 'redirecting ...')





''' Core of the IOC Container '''
import inspect

class App():
    ''' Core of the Service Container. Performs bindings and resolving
        of objects to and from the container.
    '''

    def __init__(self):
        self.providers = {}

    def bind(self, name, class_obj):
        ''' Bind classes into the container with a key value pair '''
        self.providers.update({name: class_obj})
        return self

    def make(self, name):
        ''' Retreives a class from the container by key '''
        return self.providers[name]

    def resolve(self, obj):
        ''' Takes a function or class method and resolves it's parameters
            from classes in the container
        '''

        provider_list = []
        for provider in inspect.signature(obj).parameters:
            if provider is not 'self' and provider not in inspect.getfullargspec(obj)[6]:
                provider_list.append(self.providers[provider])

        if inspect.getfullargspec(obj)[6]:
            provider_list = self.resolve_annotations(obj, provider_list)

        try:
            return obj(*provider_list)
        except TypeError:
            raise TypeError('Could not resolve the incorrect amount of objects from the container')

    def resolve_annotations(self, obj, provider_list):
        ''' Resolves class annotations (type hinting) parameters.
            Will retrieve by class type.
        '''
        for parameter in inspect.signature(obj).parameters.values():
            if parameter.annotation:
                for provider, provider_class in self.providers.items():
                    if parameter.annotation == provider_class.__class__ or parameter.annotation == provider_class:
                        provider_list.append(provider_class)

        return provider_list



''' A SassProvider Service Provider '''
from masonite.provider import ServiceProvider
from masonite.storage import Storage

class SassProvider(ServiceProvider):

    wsgi = False 
    
    def register(self):
        '''
        |--------------------------------------------------------------------------
        | Compile Sass
        |--------------------------------------------------------------------------
        |
        | Compile Sass if the libsass module is installed. Once installed, all
        | Sass files are compiled when the server is ran. This will only run
        | once when the server is first started.
        |
        '''

        Storage().compile_sass()

    def boot(self):
        pass



class Manager(object):

    def __init__(self, container=None):
        self.manage_driver = None
        self.container = container

        if container:
            self.create_driver()
    
    def load_container(self, container):
        self.container = container
        self.create_driver()
        return self

    def driver(self, driver):
        self.create_driver(driver)
        return self.container.resolve(self.manage_driver)

    def create_driver(self, driver=None):
        pass



''' Controller module '''

class Controller(object):
    ''' Base Class for Controllers '''

    def __init__(self):
        self.app = None

    def load_app(self, app):
        ''' Loads the container into the controller '''
        self.app = app
        return self



from masonite.view import view

class StackTrace(object):

    def __init__(self, trace_obj):
        self.trace_obj = trace_obj

    def render(self):
        return view('traceback', {'traceback': self.trace_obj})



import boto3

class UploadS3Driver(object):
    
    def __init__(self, Upload, StorageConfig):
        self.upload = Upload
        self.config = StorageConfig

    def store(self, fileitem, location=None):
        file_location = self.upload.driver('disk').store(fileitem)

        session = boto3.Session(
            aws_access_key_id=self.config.DRIVERS['s3']['client'],
            aws_secret_access_key=self.config.DRIVERS['s3']['secret'],
        )

        s3 = session.resource('s3')

        s3.meta.client.upload_file(
            file_location, self.config.DRIVERS['s3']['bucket'], fileitem.filename)
    def storeAs(self):
        pass



''' A WhiteNoiseProvider Service Provider '''
from masonite.provider import ServiceProvider
from whitenoise import WhiteNoise

class WhitenoiseProvider(ServiceProvider):

    wsgi = False

    def register(self):
        pass

    def boot(self, Application):
        ''' Wraps the WSGI server in a whitenoise container '''
        self.app.bind('WSGI', WhiteNoise(
            self.app.make('WSGI'), root=Application.STATIC_ROOT))

        for location, alias in self.app.make('Storage').STATICFILES.items():
            self.app.make('WSGI').add_files(location, prefix=alias)



from setuptools import setup
setup(
    name='masonite',
    packages=['masonite',
              'masonite.auth',
              'masonite.facades',
              'masonite.providers',
              'masonite.drivers',
              'masonite.managers',
             ],
    version='0.2.12.5',
    install_requires=[
        'validator.py==1.2.5',
        'cryptography==2.1.4',
        'bcrypt==3.1.4',
        'requests==2.18.4',
        'boto3==1.5.24',
    ],
    description='The core for the Masonite ',
    author='Joseph Mancuso',
    author_email='idmann509@gmail.com',
    url='https://github.com/josephmancuso/masonite',  # use the URL to the github repo
    download_url='',
    keywords=['python web framework', 'python3'],  # arbitrary keywords
    classifiers=[],
)



''' A AppProvider Service Provider '''
from masonite.provider import ServiceProvider
from masonite.request import Request
from masonite.routes import Route
from config import storage
from routes import web, api

class AppProvider(ServiceProvider):

    def register(self):
        self.app.bind('WebRoutes', web.ROUTES)
        self.app.bind('ApiRoutes', api.ROUTES)
        self.app.bind('Response', None)
        self.app.bind('Storage', storage)

    def boot(self, Environ):
        self.app.bind('Route', Route(Environ))
        self.app.bind('Request', Request(Environ).load_app(self.app))
